"""
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
"""

from typing import List
from kudubot.services.ExternalService import ExternalService


class WeatherService(ExternalService):
    """
    A service that responds with weather information for a specified location
    """

    def define_executable_file_url(self):
        """
        Defines a URL to the executable file used in this service

        :return: The URL to the executable file
        """
        return self.resolve_github_release_asset_url(
            "namboy94", "kudubot", "weather_kotlin"
        )

    def define_executable_command(self) -> List[str]:
        """
        Defines the command required to execute the executable file

        :return: The command list, for example ["java", "-jar"] for .jar files
        """
        return ["java", "-jar"]

    @staticmethod
    def define_identifier() -> str:
        """
        Defines the identifier of the service

        :return: The service's identifier
        """
        return "weather_kotlin"



"""
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import sys
import time
import logging
import argparse
import traceback
from kudubot import version
from kudubot.exceptions import InvalidConfigException
from kudubot.connections.Connection import Connection
from kudubot.config.GlobalConfigHandler import GlobalConfigHandler


# noinspection PyUnresolvedReferences
def main():  # pragma: no cover
    """
    The Main Method of the Program that starts the Connection Listener in
    accordance with the command line arguments
    """
    print("Kudubot Version " + version)

    args = parse_args()
    config_handler = GlobalConfigHandler(args.config)
    if not os.path.isdir(args.config):
        print(args.config)
        print("Config directory does not exist")
        sys.exit(1)

    # noinspection PyUnresolvedReferences
    try:

        initialize_logging(
            args.quiet,
            args.verbose,
            args.debug,
            config_handler,
            args.connection.lower()
        )
        connection = initialize_connection(
            args.connection.lower(),
            config_handler
        )

        connection.listen()

    except KeyboardInterrupt:
        print("\nBye")

    except BaseException as e:
        crashfile = os.path.join(
            config_handler.logfile_directory,
            "crashes"
        )
        with open(crashfile, 'a') as crashlog:
            crashlog.write("Crash@" + str(time.time()) + ": " + str(e))
            crashlog.write(traceback.format_exc())
        raise e


def initialize_connection(identifier: str,
                          config_handler: GlobalConfigHandler) \
        -> Connection:  # pragma: no cover
    """
    Loads the connection for the specified identifier
    If the connection was not found in the local configuration,
    the program exits.

    :param identifier: The identifier for the Connection
    :param config_handler: The config handler to use
                           to determine file paths etc.
    :return: The Connection object
    """

    try:
        config_handler.validate_config_directory()
    except InvalidConfigException as e:
        print("Loading configuration for service failed:")
        print(str(e))
        sys.exit(1)

    connections = config_handler.load_connections()
    services = config_handler.load_services()

    try:
        connection_type = list(filter(
            lambda x: x.define_identifier() == identifier,
            connections
        ))[0]
        return connection_type(services, config_handler)

    except IndexError:
        print("Connection Type " + identifier +
              " is not implemented or imported using the config file")
        sys.exit(1)
    except InvalidConfigException as e:
        print("Connection Configuration failed:")
        print(str(e))
        sys.exit(1)


def initialize_logging(quiet: bool, verbose: bool, debug: bool,
                       config_handler: GlobalConfigHandler,
                       connection_name: str):
    """
    Initializes the logging levels and files for the program.
    If neither the verbose or debug flags were provided,
    the logging level defaults to WARNING.
    Log files for ERROR, WARNING, DEBUG and INFO are always generated.
    If the size of a previous log file exceeds 1MB,
    the file is renamed and a new one is created.

    :param quiet: Can be set to disable all logging to the console.
                  Text logs are still done however.
    :param verbose: Flag that determines if the verbose mode
                    is switched on ~ INFO
    :param debug: Flag that determines if the debug mode is on ~ DEBUG
    :param config_handler: The config handler used to determine
                           the logging directory location
    :param connection_name: The name of the connection to log
    """

    stdout_handler = logging.StreamHandler(stream=sys.stdout)

    if debug:
        stdout_handler.setLevel(logging.DEBUG)
    elif verbose:
        stdout_handler.setLevel(logging.INFO)
    elif quiet:
        stdout_handler.setLevel(logging.CRITICAL)
    else:
        stdout_handler.setLevel(logging.WARNING)

    logfile = os.path.join(config_handler.logfile_directory,
                           connection_name + ".log")

    if os.path.isfile(logfile):
        if os.path.getsize(logfile) > 1000000:
            os.rename(logfile, logfile + "." + str(time.time()))

    logfile_handler = logging.FileHandler(logfile)
    logfile_handler.setLevel(logging.DEBUG)

    logging.basicConfig(
        level=logging.DEBUG,
        handlers=[stdout_handler, logfile_handler]
    )


def parse_args() -> argparse.Namespace:  # pragma: no cover
    """
    Parses the Command Line Arguments using argparse

    :return: The parsed arguments
    """

    parser = argparse.ArgumentParser()
    parser.add_argument("connection", help="The Type of Connection to use")

    parser.add_argument("-v", "--verbose", action="store_true",
                        help="Activates verbose output")
    parser.add_argument("-d", "--debug", action="store_true",
                        help="Activates debug-level logging output")
    parser.add_argument("-q", "--quiet", action="store_true",
                        help="Disables all text output")
    parser.add_argument("-c", "--config",
                        default=os.path.join(os.path.expanduser("~"),
                                             ".kudubot"),
                        help="Overrides the configuration directory location")

    return parser.parse_args()



"""
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
"""

from kudubot.config.GlobalConfigHandler import GlobalConfigHandler


class StandardConfigWriter(object):
    """
    A config handler class that writes the standard configuration
    for connections and services into the config.
    """

    def __init__(self,
                 config_handler: GlobalConfigHandler = GlobalConfigHandler()):
        """
        Initializes the Standard Config Writer's configuration file locations

        :param config_handler: The GlobalConfigHandler to use.
                               Default to the default one.
        """
        self.connection_config = \
            config_handler.global_connection_config_location
        self.service_config = config_handler.services_config_location

    def write_standard_connection_config(self):
        """
        Writes the standard connection configuration file

        :return: None
        """

        with open(self.connection_config, 'w') as config:
            for connection in [
                "@CONNECTION CliConnection",
                "@CONNECTION WhatsappConnection",
                "@CONNECTION TelegramConnection"
            ]:
                config.write(connection + "\n")

    def write_standard_service_config(self):
        """
        Writes the standard service configuration file

        :return: None
        """

        with open(self.service_config, 'w') as config:
            for service in [
                 "@NATIVE SimpleResponderService",
                 "@NATIVE ReminderService",
                 "@NATIVE AnimeReminderService",
                 "@NATIVE JokesService",
                 "@NATIVE ServiceListerService",
                 "@NATIVE AuthenticationManagerService"
                 # "@EXTERNAL helloworld_rust::HelloWorldService",
                 # "@EXTERNAL weather_kotlin::HelloKotlinService"
            ]:

                config.write(service + "\n")



"""
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
"""

from typing import Dict
from kudubot.entities.Message import Message
from kudubot.users.Contact import Contact
from kudubot.connections.Connection import Connection


class CliConnection(Connection):
    """
    The CLI kudubot connection class.
    """

    @staticmethod
    def define_identifier() -> str:
        """
        Defines the identifier for this Connection

        :return: None
        """
        return "cli"

    def send_video_message(self, receiver: Contact, video_file: str,
                           caption: str = ""):
        """
        Sending videos is not supported in the CLI

        :param receiver: The message's recipient
        :param video_file: The video file to send
        :param caption: The caption to send
        :return: None
        """
        self.logger.debug("Video messages are not supported")

    def send_audio_message(self, receiver: Contact, audio_file: str,
                           caption: str = ""):
        """
        Sending audio files is not supported in the CLI
        :param receiver: The recipient of the file
        :param audio_file: The audio file to send
        :param caption: the caption to send
        :return: None
        """
        self.logger.debug("Audio messages are not supported")

    def send_image_message(self, receiver: Contact, image_file: str,
                           caption: str = ""):
        """
        Sending image files is not supported in the CLI

        :param receiver: The recipient
        :param image_file: The image file to send
        :param caption: The caption to use
        :return: None
        """
        self.logger.debug("Image messages are not supported")

    def define_user_contact(self) -> Contact:
        """
        Specifies the User Contact, which is of no importance
        for the CLI Connection

        :return: The CLI's user Contact
        """
        return Contact(-1, "Cli", "Cli")

    def load_config(self) -> Dict[str, object]:
        """
        The CLI Connection does not require any configuration

        :return: {}
        """
        self.logger.debug("No configuration required")
        return {}

    def generate_configuration(self):
        """
        The CLI Connection does not require any configuration

        :return: None
        """
        self.logger.debug("No configuration required")

    def send_message(self, message: Message):
        """
        Prints the message to the console

        :param message: The message to 'send'
        :return: None
        """
        print("\033[93m" + message.message_body + "\033[0m")

    def listen(self):
        """
        Starts an endless loop that continuously prompts the user for input

        :return: None
        """
        while True:
            self.apply_services(
                Message("", input(), self.user_contact, self.user_contact)
            )



"""
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
"""

import requests
from bs4 import BeautifulSoup
from typing import Dict
from kudubot.entities.Message import Message
from kudubot.services.HelperService import HelperService


class JokesService(HelperService):
    """
    A Service that tells jokes
    """

    @staticmethod
    def define_identifier() -> str:
        """
        :return: The service's identifier
        """
        return "jokes"

    def define_syntax_description(self, language: str) -> str:
        """
        :param language: The language to use
        :return: The syntax description for this language
        """
        return {"en": "/joke", "de": "/witz"}[language]

    def define_help_message(self, language: str) -> str:
        """
        :param language: The language to use
        :return: The help message for the specified language
        """
        return {"en": "Tells a joke", "de": "Erzählt einen Witz"}[language]

    def determine_language(self, message: Message) -> str:
        """
        Determines the language for a message
        :param message: The message to analyze
        :return: The language that this message was in
        """
        if message.message_body.lower().strip() == "/witz":
            return "de"
        else:
            return "en"

    def define_language_text(self) -> Dict[str, Dict[str, str]]:
        """
        :return: A dictionary used for translations
        """
        return {"@response_title": {"en": "Joke", "de": "Witz"}}

    def is_applicable_to(self, message: Message):
        """
        Checks if a message is applicable to this service
        :param message: The message to analyze
        :return: True if applicable, False otherwise
        """
        return message.message_body.strip().lower() in ["/joke", "/witz"]

    def handle_message(self, message: Message):
        """
        Handles an incoming message
        :param message:
        :return:
        """
        language = self.determine_language(message)

        if language == "de":
            joke = self.load_german_joke()
        elif language == "en":
            joke = self.load_english_joke()
        else:
            joke = "ERROR_LANG_NOT_FOUND"

        self.reply(self.translate("@response_title", language), joke, message)

    # noinspection PyMethodMayBeStatic
    def load_german_joke(self) -> str:
        """
        Fetches a German joke from the internet
        :return: The joke
        """
        html = requests.get("http://witze.net/zuf%C3%A4llige-witze").text
        return BeautifulSoup(html, "html.parser").select(".joke")[0].text

    # noinspection PyMethodMayBeStatic
    def load_english_joke(self) -> str:
        """
        Fetches an English joke from the internet
        :return: The joke
        """
        html = requests.get("https://www.ajokeaday.com/jokes/random")
        soup = BeautifulSoup(html.text, "html.parser")
        return soup.select("p")[1].text



"""
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
"""

import sqlite3
import logging
from kudubot.users.Contact import Contact


# noinspection SqlDialectInspection,SqlNoDataSourceInspection,SqlResolve
class LanguageSelector(object):
    logger = logging.getLogger(__name__)
    """
    The Logger for this class
    """

    def __init__(self, db: sqlite3.Connection):
        self.db = db
        self.db.execute(
            "CREATE TABLE IF NOT EXISTS language_preferences ("
            "    user_id INTEGER CONSTRAINT constraint_name PRIMARY KEY, "
            "    lang_pref VARCHAR(255) NOT NULL,"
            "    user_initiated INTEGER NOT NULL"
            ")"
        )
        self.db.commit()
        self.logger.info("Language Selector initialized")

    def store_language_preference(self, contact: Contact, language: str,
                                  user_initiated: bool = False):
        """
        Stores the language preference for a user in the sqlite database
        :param contact: The user for which to store the preference
        :param language: The language to store
        :param user_initiated: Flag that should be set whenever the user
                               manually requests a language store
        :return: None
        """
        fetch = self.db.execute(
            "SELECT user_initiated FROM language_preferences WHERE user_id=?",
            (contact.database_id,)
        ).fetchall()

        was_user_initiated = len(fetch) > 0 and fetch[0][0]
        if was_user_initiated and not user_initiated:
            return  # Don't overwrite user defined language

        else:
            self.db.execute(
                "INSERT OR REPLACE INTO language_preferences "
                "(user_id, lang_pref, user_initiated) VALUES (?,?,?)",
                (contact.database_id, language, user_initiated)
            )
            self.db.commit()

    def get_language_preference(self, contact: Contact, default: str = "en",
                                db: sqlite3.Connection=None) -> str:
        """
        Retrieves a language from the user's preferences in the database

        :param contact: The user to check the language preference for
        :param default: A default language value used
                        in case no entry was found
        :param db: Optionally defines which database connection to use
                   (necessary for access from other thread)
        :return: The language preferred by the user
        """
        db = db if db is not None else self.db
        result = db.execute(
            "SELECT lang_pref FROM language_preferences WHERE user_id=?",
            (contact.database_id,)
        ).fetchall()
        return default if len(result) == 0 else result[0][0]



"""
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import sys
from kudubot import version
from setuptools import setup, find_packages
from kudubot.config.builder import build_external
from kudubot.config.GlobalConfigHandler import GlobalConfigHandler
from kudubot.config.StandardConfigWriter import StandardConfigWriter


def readme():
    """
    Reads the readme file and converts it to RST if pypandoc is
    installed. If not, the raw markdown text is returned
    :return: the readme file as a string
    """
    # noinspection PyBroadException
    try:
        # noinspection PyPackageRequirements,PyUnresolvedReferences
        import pypandoc
        with open("README.md") as f:
            # Convert markdown file to rst
            markdown = f.read()
            rst = pypandoc.convert(markdown, "rst", format="md")
            return rst

    except ImportError:
        # If pandoc is not installed, just return the raw markdown text
        with open("README.md") as f:
            return f.read()


def find_scripts():
    """
    Returns a list of scripts in the bin directory
    :return: the list of scripts
    """
    scripts = []

    for file_name in os.listdir("bin"):

        path = os.path.join("bin", file_name)
        if file_name == "__init__.py":
            continue
        elif not os.path.isfile(path):
            continue
        else:
            scripts.append(os.path.join("bin", file_name))

    return scripts


def run_setup():
    """
    Runs the setup method, taking care of all setuptools functionality
    :return: None
    """

    setup(name="kudubot",
          version=version,
          description="A messaging bot framework",
          long_description=readme(),
          classifiers=[
              "Environment :: Console",
              "Natural Language :: English",
              "Intended Audience :: Developers",
              "Development Status :: 4 - Beta",
              "Operating System :: POSIX :: Linux",
              "Topic :: Communications :: Chat",
              "Programming Language :: Python",
              "License :: OSI Approved :: "
              "GNU General Public License v3 (GPLv3)"
          ],
          url="https://gitlab.namibsun.net/namboy94/kudubot",
          download_url="https://gitlab.namibsun.net/namboy94/kudubot/"
                       "repository/archive.zip?ref=master",
          author="Hermann Krumrey",
          author_email="hermann@krumreyh.com",
          license="GNU GPL3",
          packages=find_packages(),
          install_requires=[
              "typing",
              "python-telegram-bot",
              "yowsup2",
              "requests",
              "bs4"],
          dependency_links=[],
          test_suite="nose.collector",
          tests_require=["nose"],
          scripts=find_scripts(),
          zip_safe=False)


def main():
    """
    Starts the setup.py script
    Writes standard config files if none exist yet
    Also attempts to build and install external services
    :return: None
    """

    if sys.argv[1] == "install":

        handler = GlobalConfigHandler()
        if not handler.validate_config_directory():
            handler.generate_configuration(False)
            StandardConfigWriter(handler).write_standard_connection_config()
            StandardConfigWriter(handler).write_standard_service_config()

        # noinspection PyBroadException
        try:
            executables = build_external()
            for executable in executables:
                os.rename(
                    executable,
                    os.path.join(
                        handler.external_services_executables_directory,
                        os.path.basename(executable)
                    )
                )
        except BaseException as e:
            print(str(e))

    run_setup()


if __name__ == "__main__":
    main()



"""
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
import sqlite3
from kudubot.users.Contact import Contact


# noinspection SqlDialectInspection,SqlNoDataSourceInspection,SqlResolve
class AddressBook(object):
    """
    Class that tracks and provides user information
    in the connection's database.

    The address book uses the following database schema:

    |id|display_name|address|
    """

    logger = logging.getLogger(__name__)
    """
    The Logger for this class
    """

    def __init__(self, database: sqlite3.Connection):
        """
        Initializes the address book. Makes sure that the
        address book's database table exists and has the correct schema

        :param database: The database connection to use
        """
        self.db = database
        self.db.execute(
             "CREATE TABLE IF NOT EXISTS address_book ("
             "    id INTEGER CONSTRAINT constraint_name PRIMARY KEY,"
             "    display_name VARCHAR(255) NOT NULL,"
             "    address VARCHAR(255) NOT NULL"
             ")"
        )
        self.db.commit()
        self.logger.info("Address Book initialized")

    def add_or_update_contact(self, contact: Contact,
                              database_override: sqlite3.Connection = None)\
            -> Contact:
        """
        Adds or updates a contact in the address book

        :param contact: The contact to insert/update
        :param database_override: Can be specified to use a different
                                  database connection, useful for calling this
                                  method from a different thread
        :return: The contact, possibly with an altered id value
                 (in case the contact was inserted, not updated)
        """
        db = self.db if database_override is None else database_override
        # Check if the contact currently exists
        old = db.execute(
            "SELECT id FROM address_book WHERE id=? OR address=?",
            (contact.database_id, contact.address)).fetchall()

        if len(old) > 0:
            db.execute(
                "UPDATE address_book SET display_name=?, address=? WHERE id=?",
                (contact.display_name, contact.address, old[0][0]))
        else:
            db.execute(
                "INSERT INTO address_book "
                "(display_name, address) VALUES (?, ?)",
                (contact.display_name, contact.address))

        db.commit()
        contact.database_id = \
            db.execute("SELECT id FROM address_book WHERE address=?",
                       (contact.address,)).fetchall()[0][0]
        return contact

    def get_contact_for_address(self, address: str,
                                database_override: sqlite3.Connection = None) \
            -> Contact:
        """
        Generates a Contact object for an address in the address book table.

        :param address: The address to look for
        :param database_override: Can be specified to use a different
                                  database connection, useful for calling this
                                  method from a different thread
        :return: The Contact object, or None if no contact was found
        """
        db = self.db if database_override is None else database_override
        result = db.execute("SELECT * FROM address_book WHERE address=?",
                            (address,)).fetchall()

        if len(result) != 1:
            # noinspection PyTypeChecker
            return None
        else:
            data = result[0]
            return Contact(int(data[0]), str(data[1]), str(data[2]))

    def get_contact_for_id(self, user_id: int,
                           database_override: sqlite3.Connection = None) \
            -> Contact:
        """
        Generates a Contact object for a user ID in the address book table

        :param user_id: The user's ID
        :param database_override: Can be specified to use a different
                                  database connection, useful for calling this
                                  method from a different thread
        :return: The user as a Contact object
        """
        db = self.db if database_override is None else database_override
        result = db.execute("SELECT * FROM address_book WHERE id=?",
                            (user_id,)).fetchall()

        if len(result) != 1:
            # noinspection PyTypeChecker
            return None
        else:
            data = result[0]
            return Contact(int(data[0]), str(data[1]), str(data[2]))



"""
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
"""

from typing import Dict
from kudubot.entities.Message import Message
from kudubot.services.HelperService import HelperService
from kudubot.services.AuthenticatedService import AuthenticatedService


class AuthenticationManagerService(HelperService, AuthenticatedService):
    """
    Lets admins manage admin rights of other users and blacklist them.
    """

    @staticmethod
    def define_identifier() -> str:
        """
        Defines the identifier for this service

        :return: The service's identifier
        """
        return "authentication_manager"

    def define_command_name(self, language: str) -> str:
        """
        Defines the command prefix for the service.
        By default, '/' followed by the service identifier from
        self.define_identifier() is used.

        :param language: The language in which to define the command name
        :return: The command name in the specified language
        """
        return {
            "en": "/auth_manage",
            "de": "/auth_verwalten"
        }[language]

    def determine_language(self, message: Message) -> str:
        """
        Determines the language of a message

        :param message: The message to check for the language
        :return: The language of the message
        """
        body = message.message_body.lower()
        for language in self.supported_languages():
            if body.startswith(self.define_command_name(language)):
                return language
        return self.get_stored_language_preference(message.sender)

    def define_language_text(self) -> Dict[str, Dict[str, str]]:
        """
        Defines the dictionary used for translating strings using the
        self.reply_translated() or self.translate() methods

        The format is:

            term: {lang: value}

        :return: The dictionary for use in translating
        """
        return {
            "@{LIST}": {"en": "list", "de": "auflisten"},
            "@{ADMIN_MAKE}": {"en": "makeadmin", "de": "adminmachen"},
            "@{BLACKLIST}": {"en": "blacklist", "de": "blacklisten"},
            "@{LIST_TITLE}": {"en": "User List", "de": "Nutzerliste"},
            "@{LIST_BODY}": {"en": "User List", "de": "Nutzerliste"},
            "@{ADMIN_MAKE_TITLE}": {"en": "Admin", "de": "Administrator"},
            "@{ADMIN_MAKE_BODY}": {"en": "The user was granted administrative "
                                         "privileges.",
                                   "de": "Dem Nutzer wurden Adminrechte "
                                         "zugeschrieben."},
            "@{BLACKLIST_TITLE}": {"en": "Blacklisted", "de": "Blacklisted"},
            "@{BLACKLIST_BODY}": {"en": "The user was blacklisted",
                                  "de": "Der Nutzer wurde geblacklisted"}
        }

    def define_help_message(self, language: str) -> str:
        """
        Defines the help message of the service in various languages

        :param language: The language to use
        :return: The help message in the language
        """
        return {
            "en": "Tool for administrators to manage blacklisted and "
                  "administrative privileges.",
            "de": "Tool für Administratoren, um blackgelistete Nutzer und"
                  "administrative Rechte zu verwalten"
        }[language]

    def define_syntax_description(self, language: str) -> str:
        """
        Defines the Syntax description of the Service in various languages

        :param language: The language in which to return the syntax message
        :return: The syntax message
        """
        return {
            "en": "/auth_manage list\n"
                  "/auth_manage makeadmin <ID>\n"
                  "/auth_manage blacklist <ID>",
            "de": "/auth_verwalten auflisten\n"
                  "/auth_verwalten adminmachen <ID>\n"
                  "/auth_verwalten blacklisten <ID>"
        }[language]

    # noinspection SqlDialectInspection,SqlDialectInspection
    # noinspection SqlNoDataSourceInspection
    def handle_message(self, message: Message):
        """
        Handles an applicable message
        Checks the mode and handles accordingly

        :param message: The message to handle
        """
        args = message.message_body.lower().split(" ")
        mode = args[1]

        if mode in [self.translate("@{LIST}", language)
                    for language in self.supported_languages()]:
            list_body = "@{LIST_BODY}:\n\n"
            results = self.connection.db.execute(
                "SELECT id, display_name, address "
                "FROM address_book").fetchall()

            for result in results:
                user_id = int(result[0])
                list_body += str(user_id) + ": " + str(result[1])
                list_body += "<" + str(result[2]) + "> "

                if self.connection.authenticator.is_admin(user_id):
                    list_body += "✅"
                elif self.connection.authenticator.is_blacklisted(user_id):
                    list_body += "❌"

                list_body += "\n"

            self.reply_translated("@{LIST_TITLE}", list_body.strip(), message)

        elif mode in [self.translate("@{ADMIN_MAKE}", language)
                      for language in self.supported_languages()]:
            user_id = int(args[2])
            self.connection.authenticator.make_admin(user_id)
            self.reply_translated(
                "@{ADMIN_MAKE_TITLE}", "@{ADMIN_MAKE_BODY}", message)

        elif mode in [self.translate("@{BLACKLIST}", language)
                      for language in self.supported_languages()]:
            user_id = int(args[2])
            self.connection.authenticator.blacklist(user_id)
            self.reply_translated(
                "@{BLACKLIST_TITLE}", "@{BLACKLIST_BODY}", message)

    def is_applicable_to(self, message: Message) -> bool:
        """
        Checks if a Message is applicable to this Service

        :param message: The message to check
        :return: True if the Message is applicable, False otherwise
        """
        language = self.determine_language(message)
        print(self.define_command_name(language))
        if not self.starts_with_command_keyword(message, language):
            return False

        body = message.message_body.lower().split(" ")

        if len(body) == 2:
            print("O")
            if body[1] == self.translate("@{LIST}", language):
                return True

        elif len(body) == 3:
            print("A")
            if body[1] in [
                self.translate("@{BLACKLIST}", language),
                self.translate("@{ADMIN_MAKE}", language)
            ]:
                try:
                    int(body[2])
                    return True
                except ValueError:
                    return False

        print("I")
        return False



"""
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
"""

import time
import telegram
import configparser
from typing import Dict, List
from kudubot.users.Contact import Contact
from kudubot.entities.Message import Message
from kudubot.connections.Connection import Connection
from kudubot.exceptions import InvalidConfigException
from kudubot.config.GlobalConfigHandler import GlobalConfigHandler


class TelegramConnection(Connection):
    """
    A Connection class that connects to Telegram
    """

    @staticmethod
    def define_identifier() -> str:
        """
        Defines the Connection's unique identifier

        :return: The Connection's identifier
        """
        return "telegram"

    def __init__(self, services: List[type],
                 config_handler: GlobalConfigHandler):
        """
        Initializes the Telegram Connection.
        Additionally to the standard Connection, a telegram.Bot object is
        generated to interface with the Telegram API

        :param services: The services to include
        :param config_handler: The GlobalConfigHandler that determines
                               the location of the configuration files
        """
        super().__init__(services, config_handler)
        self.bot = telegram.Bot(self.config["api_key"])

    def load_config(self) -> Dict[str, object]:
        """
        Loads the configuration file for the Telegram Connection
        A Telegram Connection requires only an api_key attribute in the
        [credentials] section of the config file

        Any configuration errors are raised as new InvalidConfigExceptions

        :return: The configuration, as a dictionary
        """

        try:

            self.logger.info("Parsing config at " + self.config_file_location)

            config = configparser.ConfigParser()
            config.read(self.config_file_location)
            parsed_config = dict(config.items("credentials"))

            if parsed_config["api_key"] == "":
                self.logger.warning(
                    "Config Parsing Failed. Empty API Key supplied.")
                raise InvalidConfigException(
                    "Invalid Telegram Configuration File - "
                    "Empty API key detected")

            self.logger.info("Config successfully loaded")

            return parsed_config

        except (configparser.NoSectionError,
                configparser.MissingSectionHeaderError):
            self.logger.warning("Config Parsing Failed. "
                                "No credentials section in config file.")
            raise InvalidConfigException("Invalid Telegram Configuration File "
                                         "- No credentials section")
        except KeyError as e:
            self.logger.warning("Config Parsing Failed. No attribute " +
                                str(e) + " found.")
            raise InvalidConfigException(
                "Invalid Telegram Configuration File - No '" + str(e) +
                "' attribute found")

    def generate_configuration(self):
        """
        Generates a new Configuration file for the Telegram Connection

        :return: None
        """
        with open(self.config_file_location, 'w') as config:
            config.write("[credentials]\napi_key=YourKeyHere\n")
            self.logger.info("Wrote new Configuration file at " +
                             self.config_file_location)

    def define_user_contact(self) -> Contact:
        """
        :return: The Telegram connection's contact information
        """
        return Contact(-1, "Kudubot", self.config["api_key"])

    def send_message(self, message: Message):
        """
        Sends a message using the Telegram connection

        :param message: The message to send
        :return: None
        """
        self.logger.info("Sending text Message to " +
                         message.receiver.display_name)
        self.bot.send_message(
            chat_id=message.receiver.address, text=message.message_body
        )

    def send_image_message(self, receiver: Contact, image_file: str,
                           caption: str = ""):
        """
        Sends an Image Message via Telegram

        :param receiver: The recipient of the message
        :param image_file: The image file to send
        :param caption: The caption to be displayed with the image
        :return:
        """

        with open(image_file, 'rb') as image_data:
            self.logger.info("Sending Image File to " + receiver.display_name)
            self.bot.send_photo(
                chat_id=receiver.address, photo=image_data, caption=caption)

    def send_audio_message(self, receiver: Contact, audio_file: str,
                           caption: str = ""):
        """
        Sends an Audio Message via Telegram

        :param receiver: The recipient of the message
        :param audio_file: The audio file to send
        :param caption: The caption to display together with the audio
        :return:
        """
        with open(audio_file, 'rb') as audio_data:
            self.logger.info("Sending Audio File to " + receiver.display_name)
            self.bot.send_audio(
                chat_id=receiver.address, audio=audio_data, caption=caption)

    def send_video_message(self, receiver: Contact, video_file: str,
                           caption: str = ""):
        """
        Sends a video message via Telegram

        :param receiver: The recipient of the message
        :param video_file: The video file to send
        :param caption: The caption to be displayed together with the video
        :return:
        """
        with open(video_file, 'rb') as video_data:
            self.bot.send_video(
                chat_id=receiver.address, video=video_data, caption=caption)

    def listen(self):
        """
        Listens to new Telegram messages
        Starts an infinite loop that polls for new messages every second,
        and applies the Connection's Services to these received messages.

        :return: None
        """

        try:
            update_id = self.bot.get_updates()[0].update_id
        except IndexError:
            update_id = 0

        while True:

            try:

                for update in self.bot.get_updates(
                        offset=update_id, timeout=10
                ):

                    update_id = update.update_id + 1

                    telegram_message = update.message.to_dict()
                    chat = telegram_message['chat']
                    _from = telegram_message['from']

                    name = _from["username"] \
                        if _from["username"] != "" \
                        else _from["first_name"] + _from["last_name"]

                    if chat['type'] == "group":
                        sender = self.address_book.add_or_update_contact(
                            Contact(-1, name, _from['id'])
                        )
                        group = self.address_book.add_or_update_contact(
                            Contact(-1, chat['title'], str(chat['id']))
                        )
                    else:
                        sender = self.address_book.add_or_update_contact(
                            Contact(-1, name, str(chat['id']))
                        )
                        group = None

                    message = Message(
                        "", telegram_message['text'],
                        self.user_contact, sender, group,
                        telegram_message['date']
                    )
                    self.logger.info(
                        "received message '" + message.message_body +
                        "' from " + message.sender.display_name)
                    self.apply_services(message)

                time.sleep(1)

            except telegram.error.Unauthorized:
                # The self.bot.get_update method may cause an
                # Unauthorized Error if the bot is blocked by the user
                update_id += 1

            except telegram.error.TimedOut:
                pass



"""
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
"""

# TODO Redo the external service class

import os
import json
import time
import stat
import requests
from subprocess import Popen
from bs4 import BeautifulSoup
from typing import List, Tuple, Dict
from kudubot.services.BaseService import BaseService
from kudubot.entities.Message import Message, from_dict as message_from_dict


# noinspection PyAbstractClass
class ExternalService(BaseService):
    """
    A Service which allows the use of executable files written in other
    programming languages to be used in conjunction with the kudubot framework
    """

    # noinspection PyAttributeOutsideInit
    def init(self):
        """
        Initializes the message directory for the external service and makes
        sure the executable file exists.
        If it doesn't, it will download the file from the executable file url
        specified in define_executable_file_url()
        :return: None
        """
        self.message_dir = os.path.join(
            self.connection.external_services_directory, self.identifier)
        if not os.path.isdir(self.message_dir):
            os.makedirs(self.message_dir)

        self.executable_file = os.path.join(
            self.connection.config_handler.
            external_services_executables_directory,
            self.identifier)

        if not os.path.isfile(self.executable_file):
            self.download_executable()

    def define_executable_file_url(self):
        """
        :return: An URL to the executable file
        """
        raise NotImplementedError()

    def define_executable_command(self) -> List[str]:
        """
        Defines the commands preceding the exectuable file name to
        run the program.
        For example, [python] for a python script,
        or [java, -jar] for a .jar file

        :return: The preceding commands as a list of arguments
        """
        raise NotImplementedError()

    def download_executable(self):
        """
        Downloads the executable file

        :return: None
        """
        # noinspection PyBroadException
        try:
            self.logger.info("Downloading executable file")
            with open(self.executable_file, 'wb') as destination:
                data = requests.get(self.define_executable_file_url()).content
                destination.write(data)
            self.logger.info("Download Complete")

            # Set executable permissions
            st = os.stat(self.executable_file)
            os.chmod(self.executable_file, st.st_mode | stat.S_IEXEC)

        except Exception as e:
            self.logger.error(
                "Could not download executable. Disabling Service.")
            self.logger.debug("Cause of download failure: " + str(e))
            if os.path.isfile(self.executable_file):
                os.remove(self.executable_file)

    def handle_message(self, message: Message):
        """
        Stores the Message in a json file and runs the executable file,
        then analyzes the result.

        :param message: The message to handle
        :return: None
        """
        if not os.path.isfile(self.executable_file):
            self.logger.debug("Service is disabled")
            return

        message_file, response_file = self.store_message_in_file(message)

        try:
            Popen(self.define_executable_command() +
                  [self.executable_file,
                   "handle_message", message_file, response_file,
                   self.connection.database_file_location]).wait()

            response = self.load_json(response_file)

            if response["mode"] == "reply":
                self.connection.send_message(
                    self.retrieve_message_from_file(message_file))

        except BaseException as e:
            self.logger.error(
                "Execution of external Service failed: " + str(e))

        self.cleanup(message_file, response_file)

    def is_applicable_to(self, message: Message) -> bool:
        """
        Writes the message to a JSON file, then communicates with the
        executable to check if the message is applicable to the file

        :param message: The message to analyze
        :return: None
        """
        if not os.path.isfile(self.executable_file):
            self.logger.debug("Service is disabled")
            return False

        message_file, response_file = self.store_message_in_file(message)
        applicable = False

        try:
            Popen(self.define_executable_command() +
                  [self.executable_file,
                   "is_applicable_to", message_file, response_file,
                   self.connection.database_file_location]).wait()

            response = self.load_json(response_file)

            applicable = bool(response["is_applicable"])

        except BaseException as e:
            self.logger.error(
                "Execution of external Service failed: " + str(e))

        self.cleanup(message_file, response_file)
        return applicable

    def store_message_in_file(self, message: Message) -> Tuple[str, str]:
        """
        Stores a message in a json file.
        The filename of the file will be the current time.
        Also generates a response file location in which the executable may
        write a response into

        :param message: The message to save
        :return: The location of the stored message json file,
                 the location of the response file
        """

        json_data = message.to_dict()

        while True:  # Make sure that file does not exist
            message_file = os.path.join(self.message_dir, str(time.time()))
            if not os.path.isfile(message_file):
                with open(message_file + ".json", 'w') as json_file:
                    json.dump(json_data, json_file)
                return message_file + ".json", message_file + "-response.json"

    # noinspection PyMethodMayBeStatic
    def load_json(self, response_file: str) -> Dict[str, type]:
        """
        Loads a json file into a dictionary

        :param response_file: The file to parse
        :return: The generated dictionary
        """

        with open(response_file, 'r') as f:
            content = json.load(f)

        return content

    def retrieve_message_from_file(self, message_file: str) -> Message:
        """
        Loads a message from a message json file

        :param message_file: The json file location
        :return: The generates Message object
        """
        return message_from_dict(self.load_json(message_file))

    # noinspection PyMethodMayBeStatic
    def cleanup(self, message_file: str, response_file: str):
        """
        Deletes the message and response files

        :param message_file: The message file to delete
        :param response_file: The response file to delete
        :return: None
        """

        if os.path.isfile(message_file):
            os.remove(message_file)
        if os.path.isfile(response_file):
            os.remove(response_file)

    # noinspection PyMethodMayBeStatic
    def resolve_github_release_asset_url(self, owner: str, repository: str,
                                         filename: str) -> str:
        """
        Generates a download URL for a Github release artifact/asset

        :param owner: The owner of the Github repository
        :param repository: The name of the repository
        :param filename: The file to search for
        :return: The URL to the file, if it was found.
                 If not, an empty string is returned
        """

        release_page = "https://github.com/" + owner + "/" + \
                       repository + "/releases/latest"
        soup = BeautifulSoup(requests.get(release_page).text, "html.parser")
        downloads = soup.select(".release-downloads")[0]

        for download in downloads.select("a"):
            name = download.select("strong")[0].text

            if name == filename:
                return "https://github.com" + download["href"]

        return ""
